(** Provides a SAX-like interface, where the parser invokes user callbacks as it processes a document. *)

open Yaxpo

(** {C {b Definition of callbacks}} *)

(** Each field should be set to a function that will be invoked when the corresponding token is parsed from the XML data.

The callbacks are invoked in {e continuation-passing style} (CPS). They are each passed an additional argument besides what you expect, a procedure of type [(unit->unit)]. When your callback is finished processing, instead of returning control, it should issue a tail-call to this procedure. *)
type xml_callbacks =
    {
      start_element : (qname -> att list -> (unit->unit) -> unit);
      characters : (string -> (unit->unit) -> unit);
      cdata : (string -> (unit->unit) -> unit);
      comment : (string -> (unit->unit) -> unit);
      pi : (string -> string -> (unit->unit) -> unit);
      end_element : (qname -> (unit->unit) -> unit);
    }

(** {C {b Parsing routines}} *)

(** Parse the next element from the reader. Accepts a [Cps_reader.t], your callbacks, and a continuation. In most cases, you can use [(fun () -> ())] for the continuation.

@raise U_char.Bad_char if an invalid character is encountered.
@raise Utf8_char.Bad_encoding if a UTF-8 decoding error occurs.
@raise Lex_error if a lexical error is encountered.
@raise Parse_error if a parsing error is encountered.*)
val parse_element : (#Cps_reader.t -> xml_callbacks -> (unit->unit) -> unit)

(** Parse a document (consisting of some comments and PIs followed by a document element)

@raise U_char.Bad_char if an invalid character is encountered.
@raise Utf8_char.Bad_encoding if a UTF-8 decoding error occurs.
@raise Lex_error if a lexical error is encountered.
@raise Parse_error if a parsing error is encountered, or if content comes in the wrong order. *)
val parse_document : (#Cps_reader.t -> xml_callbacks -> (unit->unit) -> unit)

(**/**)

(* internal use *)
val parse_rest_of_element : (#Cps_reader.t -> xml_callbacks -> qname -> (unit->unit) -> unit)
